package controller;

import model.World;
import model.elements.Ant;
import model.factory.PopulationFactory;
import model.miscellaneous.RND;
import model.types.Point;
import view.IView;

import java.util.*;
import java.util.stream.Collectors;

import static model.factory.PopulationFactory.generatePopulation;

/**
 * Created by filip on 14.04.16..
 */
public class ACOalgorithm {

    private IView view;

    private World world;

    public Thread acoThread;

    private double ALPHA             = 1;
    private int    MINIMAL_SOLUTIONS = 2;

    public ACOalgorithm(IView view, World world) {
        this.view = view;
        this.world = world;
        this.acoThread = new Thread(new Runnable() {
            @Override
            public void run() {
                start();
            }
        });

    }

    public void start() {
        int iter = 0;
        while (true) {

//            List<Ant> population = PopulationFactory.generatePopulation(World.NUMBER_OF_AGENTS, world.getNest());
            iter++;
            int solutionsFound = 0;
            List<Set<Point>> solutions = new LinkedList<>();
            while (solutionsFound<World.NUMBER_OF_SOLUTIONS) {

                // one iteration of algorithm
                for (Ant a : world.getPopulation()) {
                    Point nextp = getNextPoint(world.getNeighbours(a.getCurrentPosition()), a);

                    a.move(nextp);   // PROVJERITI!
                    if (world.getElement(a.getCurrentPosition()).getFoodLevel() > 0) {
                        //foodNotFound = false;
                        //updatePheromones(a);                                                // PROVJERITI!
                        solutionsFound++;
                        solutions.add(a.getHistory());
                        a.goHome();
                    }
                    //if (a.getHistory().size() > 10 * world.getWidth() * world.getHeight()) a.goHome();
                }

                // draw state of colony
                view.produce();


                //slow down
               /* try {
                    Thread.sleep(20);
                } catch (InterruptedException e) {
                    e.printStackTrace();
                }
*/

                }

                solutions = solutions.stream().sorted((a,b) -> a.size()-b.size()).collect(Collectors.toList());
                for(int i = 0; i<World.BEST_SOLUTIONS; ++i){
                    updatePheromones(solutions.get(i));
                }

                // food found, update pheromone trail and reset population
                world.resetPopulation(World.NUMBER_OF_AGENTS);
                world.evaporatePheromones();
            System.out.println(iter);
            }
        }


    private void updatePheromones(Ant ant){
        double d = ant.getCurrentPosition().distance(world.getNest());
        double increase = (d/ant.getHistory().size());
        ant.getHistory().forEach(p -> world.getElement(p).increasePheromone(increase));
    }

    private void updatePheromones(Set<Point> s){
        double d = s.size();
        for(Point p : s){
            world.getElement(p).increasePheromone(world.getWidth()/d);
        }
    }

    private void evaporatePheromones() {
        for (int i = 0, width = world.getWidth(); i < width; ++i) {
            for (int j = 0, height = world.getHeight(); j < height; ++j) {
                world.getElement(i,j).decreasePheromon();
            }
        }
    }

    private Point getNextPoint(List<Point> neighbours, Ant ant) {

        //List<Point> filteredNeighbours = neighbours.stream().filter(n -> !ant.getHistory().contains(n)).collect(Collectors.toList());
        for(Point p : neighbours){
            if(world.getElement(p).getFoodLevel()>0){
                return p;
            }
        }

        neighbours = neighbours.stream().filter(point -> !point.equals(ant.getPreviousPosition())).collect(Collectors.toList());
        if (neighbours.isEmpty()) return neighbours.get(RND.getInstance().nextInt(neighbours.size()));

        neighbours = neighbours.stream().filter(p -> !p.equals(world.getNest())).collect(Collectors.toList());

        double pheromoneSum = neighbours.stream().mapToDouble(p -> Math.pow(world.getElement(p).getFoodPheromoneLevel(), World.ALPHA)).sum();
        //if(pheromoneSum<1E-3) return neighbours.get(RND.getInstance().nextInt(neighbours.size()));

        double sum = 0;
        double rand = Math.random();

        for (Point n : neighbours) {
            sum += Math.pow(world.getElement(n).getFoodPheromoneLevel(), World.ALPHA) / pheromoneSum;
            if (rand < sum) return n;

        }
        return neighbours.get(neighbours.size() - 1);
    }

    public void set() {

    }
/*
        if(Math.abs(pheromoneSum)<1E-3) return neighbours.get(RND.getInstance().nextInt(neighbours.size()));
        neighbours.sort((p1,p2) ->
                Double.compare(
                        world.getElement(p2).getFoodPheromoneLevel(),
                        world.getElement(p1).getFoodPheromoneLevel()
                )
        );
        return neighbours.get(0); */

 /*       double pheromoneSum = neighbours.stream().mapToDouble(p -> world.getElement(p).getFoodPheromoneLevel()).sum();
        int randomPick = RND.getInstance().nextInt(1000);
        int cummulativeSum = 0;

        if(Math.abs(pheromoneSum)<1E-3) return neighbours.get(RND.getInstance().nextInt(neighbours.size()));

        for(int i = 0; i<neighbours.size();++i){
            cummulativeSum = (i == neighbours.size()-1) ? 1000 : cummulativeSum+(int)(world.getElement(neighbours.get(i)).getFoodPheromoneLevel()/pheromoneSum)*1000;
            if(randomPick <= cummulativeSum) {
                if(i<neighbours.size()-1) {
                    return neighbours.get(i);
                } else{
                    return neighbours.get(i);
                }
            }

        }

        return null;*/
}

