package model.elements;

import model.World;
import model.types.CellType;

/**
 * Created by filip on 08.04.16..
 */
public class Cell {
    private static final double EVAPORATE_COEFFICIENT = 0.5;
    private CellType type;
    private double homePheromoneLevel;
    private double foodPheromoneLevel;
    public static double pherMax = 0.1;
    private int foodSupplies;

    public Cell() {
        this(CellType.FREE);
    }

    public Cell(CellType type){
        this.type = type;
        foodPheromoneLevel = 1E-3;
    }

    public CellType getType () {
        return type;
    }
    public void setWall() {
        type = CellType.WALL;
        resetProperties();
    }

    public void setNest() {
        type = CellType.NEST;
        resetProperties();
    }

    public void setFree() {
        type = CellType.FREE;
        resetProperties();
    }

    public boolean isWall() {
        return type == CellType.WALL;
    }

    public boolean isFree() {
        return type == CellType.FREE;
    }

    public boolean isNest() {
        return type == CellType.NEST;
    }

    public int getFoodLevel() {
        return foodSupplies;
    }

    public void setFoodLevel(int level) {this.foodSupplies = level;}

    public double getHomePheromoneLevel() {
        return homePheromoneLevel;
    }


    public double getFoodPheromoneLevel() {
        return foodPheromoneLevel;
    }

    public void increasePheromone(double k){
        foodPheromoneLevel += k;
        if(foodPheromoneLevel > pherMax){
            pherMax = 2*foodPheromoneLevel;
        }
    }

    public void decreasePheromon(){
        homePheromoneLevel*= World.EVAPORATION_RATE;
        foodPheromoneLevel*=World.EVAPORATION_RATE;
    }

    public void setPheromoneLevel(Double d){
        foodPheromoneLevel = d;
    }

    private void resetProperties() {
        homePheromoneLevel = 0;
        foodPheromoneLevel = 1E-3;
        foodSupplies = 0;
    }

    @Override
    public String toString() {
       switch(type){
           case FREE: if(foodSupplies>9) return "*";
                      else if(foodSupplies>0) return Integer.toString(foodSupplies);
                      else return " ";
           case WALL: return "#";
           case NEST: return Integer.toString(foodSupplies);
       }
        return null;
    }
}
