package model.factory;

import model.World;
import model.miscellaneous.RND;
import model.elements.Cell;
import model.types.DomainType;
import model.types.Point;
import view.GraphicView;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.LinkedList;
import java.util.List;

/**
 * Created by filip on 08.04.16..
 */
public class MazeFactory {

    public static Cell[][] parseMazeFromList(List<String> input, Point food, Point nest){

        List<Cell[]> l = new LinkedList<>();
        Cell[][] maze = null;

        int cnt = 0;
        for(String line : input){
            Cell[] c = new Cell[line.length()];
            for(int i = 0, n = line.length(); i<n;++i){
                c[i] = new Cell();
                switch(line.charAt(i)){
                    case 'n':
                        c[i].setNest();
                        nest.setX(i);
                        nest.setY(cnt);
                        break;
                    case 'f':
                        c[i].setFoodLevel(10);
                        food.setX(i);
                        food.setY(cnt);
                        break;
                    case '#':
                        c[i].setWall();
                        break;
                }
            }
            l.add(c);
            ++cnt;


        }

        maze = new Cell[l.size()][];
        for(int i = 0, n = l.size(); i<n; ++i){
            maze[i] = l.get(i);
        }


        return maze;
    }

    public static Cell[][] parseMazeFromFile(Path path, Point food, Point nest){

        List<Cell[]> l = new LinkedList<>();
        Cell[][] maze = null;

        try(BufferedReader reader = new BufferedReader(new InputStreamReader(new FileInputStream(path.toFile())))){
            String line;
            int cnt = 0;
            while((line = reader.readLine())!=null){
                Cell[] c = new Cell[line.length()];
                for(int i = 0, n = line.length(); i<n;++i){
                    c[i] = new Cell();
                    switch(line.charAt(i)){
                        case 'n':
                            c[i].setNest();
                            nest.setX(i);
                            nest.setY(cnt);
                            break;
                        case 'f':
                            c[i].setFoodLevel(10);
                            food.setX(i);
                            food.setY(cnt);
                            break;
                        case '#':
                            c[i].setWall();
                            break;
                    }
                }
                l.add(c);
                ++cnt;
            }

            maze = new Cell[l.size()][];
            for(int i = 0, n = l.size(); i<n; ++i){
                maze[i] = l.get(i);
            }


        } catch(IOException e){

        }

        return maze;
    }

    public static Cell[][] schoolExampleMaze(DomainType type, Point nest, Point food){

        int size = type.getSize();
        int width = type.getWidth();

        Cell[][] maze = emptyMaze(size);

        drawLine(size/2-10, 20, 2, 30, maze);

        return maze;
    }

    /**
     *
     * @param rowStart
     * @param colStart
     * @param direction  1=NORTH, 2=EAST, 3=SOUTH, 4=WEST
     * @param size
     */
    private static void drawLine(int rowStart, int colStart, int direction, int size, Cell[][] maze){
        if(direction == 1){
            for(int i = 0; i<size; ++i){
                maze[rowStart--][colStart].setWall();
            }
        } else if(direction == 2){
            for(int i = 0; i<size; ++i){
                maze[rowStart][colStart++].setWall();
            }
        } else if(direction == 3){
            for(int i = 0; i<size; ++i){
                maze[rowStart++][colStart].setWall();
            }
        } else if(direction == 4){
            for(int i = 0; i<size; ++i){
                maze[rowStart][colStart--].setWall();
            }
        }
    }


    public static Cell[][] generateSimpleMaze(DomainType type, Point nest, Point food){

        int size = type.getSize();
        int width = type.getWidth();

        Cell[][] maze = emptyMaze(size);
        addFrame(maze, size);

        for(int i = size/4; i<=3*size/4;++i){
            maze[i][size/2].setWall();
/*
        maze[3*size/4-10][3].setNest();
        nest.setX(3);
        nest.setY(3*size/4-10);

        maze[3*size/4-10][size - 4].setFoodLevel(100);
        food.setX(size-4);
        food.setY(3*size/4-10);
*/

        }
        maze[size/2][3].setNest();
        nest.setX(3);
        nest.setY(size/2);

        maze[size/2][size - 4].setFoodLevel(100);
        food.setX(size-4);
        food.setY(size/2);
        return maze;
    }

    public static Cell[][] generateMaze(DomainType type, Point nest, Point food) {

        int size = type.getSize();
        int width = type.getWidth();

        Cell[][] maze = emptyMaze(size);

        addFrame(maze, size);
        addStripes(maze, size, width);

        nest.setX((RND.getInstance().nextInt(size) / width) * width + 1);
        nest.setY((RND.getInstance().nextInt(size) / width) * width + 1);

        recursiveLabyrinth(nest, new boolean[size][size], maze, width, size);

        FoodFactory.generateRandomFood(maze, food, 1);

        return maze;
    }

    private static Cell[][] emptyMaze(int size) {
        Cell[][] maze = new Cell[size][size];
        for (int i = 0; i < size; ++i) {
            for (int j = 0; j < size; ++j) {
                maze[j][i] = new Cell();
            }
        }
        return maze;
    }

    private static void addFrame(Cell[][] maze, int size) {

        for (int i = 0; i < size; ++i) {
            maze[0][i].setWall();
            maze[size - 1][i].setWall();
            maze[i][0].setWall();
            maze[i][size - 1].setWall();
        }

    }

    private static void addStripes(Cell[][] maze, int size, int width) {
        for (int i = width; i < size - 2; i += width) {
            for (int j = 0; j < size; j++) {
                maze[j][i].setWall();
                maze[i][j].setWall();
            }
        }
    }



        private static void recursiveLabyrinth(Point p, boolean[][] v, Cell[][] maze, int complexity, int size) {

        List<Point> neighbours = getNeighbours(p, maze, size, complexity);

        v[p.getY()][p.getX()] = true;

        while (neighbours.size() > 0) {

            Point neighbour = neighbours.remove(RND.getInstance().nextInt(neighbours.size()));

            int x = p.getX();
            int y = p.getY();
            int xNext = neighbour.getX();
            int yNext = neighbour.getY();

            if (!v[yNext][xNext]) {

                if (yNext == y) { // u istom redku
                    int i = (xNext + x) / 2 + complexity / 2 - (complexity + 1) % 2;
                    for (int j = yNext; j < yNext + complexity - 1 && j < size - 1; ++j) {
                        maze[j][i].setFree();
                    }
                } else { // u istom stubcu
                    int j = (yNext + y) / 2 + complexity / 2 - (complexity + 1) % 2;
                    for (int i = xNext; i < x + complexity - 1 && i < size - 1; ++i) {
                        maze[j][i].setFree();
                    }
                }

                recursiveLabyrinth(new Point(xNext, yNext), v, maze, complexity, size);
            }
        }
    }

    private static List<Point> getNeighbours(Point p, Cell[][] maze, int size, int complexity) {

        List<Point> neighbours = new ArrayList<>();

        int x = p.getX();
        int y = p.getY();

        if (x - complexity > 0) {
            neighbours.add(new Point(x - complexity, y));
        }
        if (x + complexity < size - 1) {
            neighbours.add(new Point(x + complexity, y));
        }
        if (y - complexity > 0) {
            neighbours.add(new Point(x, y - complexity));
        }
        if (y + complexity < size - 1) {
            neighbours.add(new Point(x, y + complexity));
        }

        return neighbours;
    }
}
