package view;

import model.World;
import model.elements.Ant;
import model.elements.Cell;
import model.factory.MazeFactory;
import model.types.*;

import javax.swing.*;
import javax.swing.border.Border;
import javax.swing.border.TitledBorder;

import java.awt.*;
import java.awt.Point;
import java.awt.event.ActionEvent;
import java.io.*;
import java.util.*;

/**
 * Created by filip on 12.05.16..
 */
public class GraphicView implements IView {

    private static final Rectangle freeBlock = new Rectangle();

    private int width;
    private int height;
    private JFrame mainFrame;
    private MyPanel panel;
    private JPanel sidePanel1;
    private JPanel sidePanel2;
    private JPanel sidePanel;
    private JPanel bottomPanel;
    private JPanel masterPanel;
    private World world;

    private Thread thread;

    private IntegerInputComponent populationSize;
    private IntegerInputComponent solutionsSize;
    private IntegerInputComponent bestSolutionsSize;
    private DoubleInputComponent alphaFactor;
    private DoubleInputComponent evaporationRate;

    private final static int GRID_ELEMENT_SIZE = 6;

    public GraphicView(World world){



        this.world = world;
        this.width = GRID_ELEMENT_SIZE*world.getWidth();
        this.height = GRID_ELEMENT_SIZE*world.getHeight();
        mainFrame = new JFrame("ACO simulator");
        panel = new MyPanel();
        masterPanel = new JPanel();
        masterPanel.add(panel);
        //sidePanel = new JPanel(new GridLayout(2,1,0,10));
        sidePanel = new JPanel(new BorderLayout());
        sidePanel.setPreferredSize(new Dimension(200, masterPanel.getHeight()));
        sidePanel1 = new JPanel(new GridLayout(8,1,0,20));

        JPanel sidePanel22 = new JPanel(new GridLayout(4,1,0,10));
        sidePanel2 = new FileChooser();

        sidePanel22.add(new JButton(new AbstractAction("Choose file") {
            @Override
            public void actionPerformed(ActionEvent e) {
                thread.suspend();
                JFileChooser fc = new JFileChooser();
                int ret = fc.showOpenDialog(mainFrame);
                if (ret == JFileChooser.APPROVE_OPTION){
                    File file = fc.getSelectedFile();
                    try {
                        parse(file);
                    } catch (Exception ex){
                        thread.resume();
                    }
                }
            }

            private void parse(File file){
                try(BufferedReader reader = new BufferedReader(new InputStreamReader(new FileInputStream(file)))){
                    int numberOfAgents = Integer.parseInt(reader.readLine());
                    int numberOfSolutions = Integer.parseInt(reader.readLine());
                    int bestSolutions = Integer.parseInt(reader.readLine());
                    double alpha = Double.parseDouble(reader.readLine());
                    double evaporate = Double.parseDouble(reader.readLine());

                    java.util.List<String> input = new LinkedList<String>();
                    String line;
                    while ((line = reader.readLine())!=null){
                        input.add(line);
                    }

                    model.types.Point nest = new model.types.Point(0,0);
                    model.types.Point food = new model.types.Point(0,0);
                    Cell[][] maze = MazeFactory.parseMazeFromList(input, food, nest);
                    world.reMake(maze, food, nest);
                    World.NUMBER_OF_SOLUTIONS = numberOfSolutions;
                    World.NUMBER_OF_AGENTS = numberOfAgents;
                    World.BEST_SOLUTIONS = bestSolutions;
                    World.ALPHA = alpha;
                    World.EVAPORATION_RATE = evaporate;
                    GraphicView.this.populationSize.setValue(numberOfAgents);
                    GraphicView.this.solutionsSize.setValue(numberOfSolutions);
                    GraphicView.this.bestSolutionsSize.setValue(bestSolutions);
                    GraphicView.this.alphaFactor.setValue(alpha);
                    GraphicView.this.evaporationRate.setValue(evaporate);
                    mainFrame.revalidate();
                    mainFrame.repaint();
                } catch (IOException ex){
                    throw new IllegalArgumentException();
                }

            }
        }));

        sidePanel22.add(new JButton(new AbstractAction("Random hard") {
            @Override
            public void actionPerformed(ActionEvent e) {
                thread.suspend();
                model.types.Point food = new model.types.Point(0,0);
                model.types.Point nest = new model.types.Point(0,0);
                Cell[][] maze = MazeFactory.generateMaze(DomainType.MEDIUM_HARD, nest, food);
                world.reMake(maze, food, nest);
                mainFrame.revalidate();
                mainFrame.pack();
                mainFrame.repaint();
            }
        }));

        sidePanel22.add(new JButton(new AbstractAction("Random simple") {
            @Override
            public void actionPerformed(ActionEvent e) {
                thread.suspend();
                model.types.Point food = new model.types.Point(0,0);
                model.types.Point nest = new model.types.Point(0,0);
                Cell[][] maze = MazeFactory.generateMaze(DomainType.MEDIUM_NORMAL, nest, food);
                world.reMake(maze, food, nest);
                mainFrame.revalidate();
                mainFrame.pack();
                mainFrame.repaint();
            }
        }));

        sidePanel22.add(new JButton(new AbstractAction("Reset current") {
            @Override
            public void actionPerformed(ActionEvent e) {
                thread.suspend();
                world.reset();
                mainFrame.revalidate();
                mainFrame.pack();
                mainFrame.repaint();

            }
        }));
      //  sidePanel22.setPreferredSize(new Dimension(200,400));

       // sidePanel22.add(new JButton("Random simple"));
       // sidePanel22.add(new JButton("Random hard"));

        masterPanel.add(sidePanel);
        bottomPanel = new JPanel(new GridLayout(1,2));
        populationSize      = new IntegerInputComponent("Population size",    World.NUMBER_OF_AGENTS, n -> World.NUMBER_OF_AGENTS = n);
        solutionsSize       = new IntegerInputComponent("Solutions size",     World.NUMBER_OF_SOLUTIONS, n -> World.NUMBER_OF_SOLUTIONS = n);
        bestSolutionsSize   = new IntegerInputComponent ("Best solutions",     World.BEST_SOLUTIONS, n -> World.BEST_SOLUTIONS = n);
        alphaFactor         = new DoubleInputComponent ("Convergence factor", World.ALPHA, n -> World.ALPHA = n);
        evaporationRate     = new DoubleInputComponent ("Evaporation rate",   World.EVAPORATION_RATE, n -> World.EVAPORATION_RATE = n);



        sidePanel1.setBorder(BorderFactory.createTitledBorder(BorderFactory.createLineBorder(Color.BLACK, 2), "Parameters"));

        sidePanel1.add(populationSize);
        sidePanel1.add(solutionsSize);
        sidePanel1.add(bestSolutionsSize);
        sidePanel1.add(alphaFactor);
        sidePanel1.add(evaporationRate);
        sidePanel1.add(new JButton(new AbstractAction("Change") {
            @Override
            public void actionPerformed(ActionEvent e) {

                Integer i = null;
                try{
                    i = Integer.parseInt(populationSize.getValue());
                } catch (NumberFormatException ex){
                    populationSize.setValue(World.NUMBER_OF_AGENTS);
                }
                if(i != null && i>=1){
                    World.NUMBER_OF_AGENTS = i;
                } else {
                    populationSize.setValue(World.NUMBER_OF_AGENTS);
                }

                i = null;
                try{
                    i = Integer.parseInt(solutionsSize.getValue());

                } catch (NumberFormatException ex){
                    solutionsSize.setValue(World.NUMBER_OF_SOLUTIONS);
                }
                if(i != null && i>0 && i<=World.NUMBER_OF_AGENTS){
                    World.NUMBER_OF_SOLUTIONS = i;
                } else {
                    solutionsSize.setValue(World.NUMBER_OF_SOLUTIONS);
                }


                i = null;
                try{
                    i = Integer.parseInt(bestSolutionsSize.getValue());
                } catch (NumberFormatException ex){
                    bestSolutionsSize.setValue(World.BEST_SOLUTIONS);
                }
                if(i != null && i>0 && i<=World.NUMBER_OF_SOLUTIONS){
                    World.BEST_SOLUTIONS = i;
                } else {
                    bestSolutionsSize.setValue(World.BEST_SOLUTIONS);
                }

                Double d = null;
                try{
                    d = Double.parseDouble(alphaFactor.getValue());
                } catch (NumberFormatException ex){
                    alphaFactor.setValue(World.ALPHA);
                }
                if(d!=null && d>0){
                    World.ALPHA = d;
                } else {
                    alphaFactor.setValue(World.ALPHA);
                }

                d = null;
                try{
                    d = Double.parseDouble(evaporationRate.getValue());
                } catch (NumberFormatException ex){
                    evaporationRate.setValue(World.EVAPORATION_RATE);
                }
                if(d!=null && d>0){
                    World.EVAPORATION_RATE = d;
                } else {
                    evaporationRate.setValue(World.EVAPORATION_RATE);
                }
            }
        }));

        //sidePanel2.add(new , "Select domain");
        //sidePanel2.setBorder(new TitledBorder(BorderFactory.createLineBorder(Color.BLACK, 2), "Domain"));
        JPanel tempDataPanel = new JPanel();
        //
        tempDataPanel.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));

        sidePanel.add(sidePanel1);
        sidePanel22.setBorder(new TitledBorder(BorderFactory.createLineBorder(Color.BLACK, 2), "Domain model"));
        //sidePanel22.setLayout(new GridLayout(4,1));
     //   sidePanel22.setPreferredSize(new Dimension(200, 100));
        sidePanel.add(sidePanel22, BorderLayout.SOUTH);

        bottomPanel = new JPanel(new GridLayout(1,2));
        bottomPanel.add(new JButton(new AbstractAction("Start") {
            @Override
            public void actionPerformed(ActionEvent e) {
                if(!thread.isAlive()) thread.start();
                else {
                    thread.resume();
                }
            }
        }));

        bottomPanel.add(new JButton(new AbstractAction("Pause") {
            @Override
            public void actionPerformed(ActionEvent e) {
                if(thread.isAlive()) thread.suspend();
            }
        }));

        SwingUtilities.invokeLater(() -> {
            mainFrame.setDefaultCloseOperation(WindowConstants.EXIT_ON_CLOSE);
            mainFrame.setSize(width*GRID_ELEMENT_SIZE,height*GRID_ELEMENT_SIZE);

            mainFrame.setLayout(new BorderLayout());

            mainFrame.add(masterPanel);
            mainFrame.add(sidePanel, BorderLayout.EAST);
            mainFrame.add(bottomPanel, BorderLayout.SOUTH);
            mainFrame.pack();
            //mainFrame.setResizable(false);
            mainFrame.setVisible(true);
        });
    }

    public void produce(){
        panel.repaint();
    }


    class MyPanel extends JPanel {
        int i = 0;

        @Override
        public Dimension getPreferredSize() {
            return new Dimension(world.getWidth()*GRID_ELEMENT_SIZE, world.getHeight()*GRID_ELEMENT_SIZE);
        }

        @Override
        public void paint(Graphics g) {
            super.paint(g);
            Color[] colors = {Color.RED, Color.GREEN, Color.BLUE };
            g.setColor(colors[i++%colors.length]);
           // g.fillRect(10,10,30,30);

           for(int i = 0, h = world.getWidth(); i<h; ++i){
                for(int j = 0, w = world.getHeight(); j<w; ++j){
                    switch(world.getElement(i,j).getType()) {
                        case FREE:
                            g.setColor(Color.WHITE);
                            break;
                        case WALL:
                            g.setColor(Color.BLACK);
                            break;
                        case NEST:
                            g.setColor(Color.RED);
                            break;
                    }
                    int food = world.getElement(i,j).getFoodLevel();
                    double pheromone = world.getElement(i,j).getFoodPheromoneLevel();
                    if(pheromone!=0 && world.getElement(i,j).getType() == CellType.FREE) {
                        //System.out.println(pheromone);
                        g.setColor(new Color(0.0f,1.0f,0.0f, (float)(pheromone/Cell.pherMax)));
                        //g.setColor(new Color(0,255,0,(int)Math.min(pheromone*255,255)));
                    }
                    if(food!=0) g.setColor(Color.BLACK);

                    g.fillRect(i*GRID_ELEMENT_SIZE, j*GRID_ELEMENT_SIZE, GRID_ELEMENT_SIZE, GRID_ELEMENT_SIZE);
                    }
                }

            for(Ant ant : world.getPopulation()){
                int dx = ant.getCurrentPosition().getX()-ant.getPreviousPosition().getX();
                int dy = ant.getCurrentPosition().getY()-ant.getPreviousPosition().getY();
                int x = ant.getPreviousPosition().getX()*GRID_ELEMENT_SIZE;
                int y = ant.getPreviousPosition().getY()*GRID_ELEMENT_SIZE;
                for(int k = 1; k<=GRID_ELEMENT_SIZE; ++k){
                    g.setColor(Color.WHITE);
                    g.fillOval(x,y,GRID_ELEMENT_SIZE,GRID_ELEMENT_SIZE);
                    g.setColor(Color.PINK);
                    x+=dx;
                    y+=dy;
                    g.fillOval(x,y,GRID_ELEMENT_SIZE,GRID_ELEMENT_SIZE);
                }
                }
            }

    }

    public void setThread(Thread t){
        thread = t;
    }
    }

